import pytest

from dsenum.permutation_group import (
    DerivativeStructurePermutation,
    is_permutation_group,
)
from dsenum.superlattice import generate_symmetry_distinct_superlattices
from dsenum.utils import get_lattice


@pytest.mark.parametrize("kind", ["fcc", "bcc", "sc", "hex", "tet", "hcp"])
def test_permutations(kind):
    structure = get_lattice(kind)
    frac_coords = structure.frac_coords
    for index in range(1, 10 + 1):
        list_reduced_HNF, rotations, translations = generate_symmetry_distinct_superlattices(
            index, structure, return_symops=True
        )
        for hnf in list_reduced_HNF:
            dsperm = DerivativeStructurePermutation(hnf, frac_coords, rotations, translations)
            assert is_permutation_group(dsperm.prm_t)
            prm_all = dsperm.get_symmetry_operation_permutations()
            assert is_permutation_group(prm_all)



import numpy as np
import pytest
from pymatgen.core import Lattice, Structure

from dsenum import StructureEnumerator, ZddStructureEnumerator
from dsenum.utils import get_lattice


@pytest.mark.parametrize("method", ["direct", "zdd", "zdd_count"])
@pytest.mark.benchmark(group="basic-binary")
def test_basic_binary(method, benchmark):
    setting = {
        "base_structure": get_lattice("fcc"),
        "index": 16,
        "num_types": 2,
        "remove_incomplete": True,
        "remove_superperiodic": True,
    }

    if method == "direct":
        se = StructureEnumerator(**setting)
        benchmark.pedantic(se.generate, kwargs={"output": "poscar"}, iterations=1)
    else:
        zse = ZddStructureEnumerator(**setting)
        if method == "zdd":
            benchmark.pedantic(zse.generate, kwargs={"output": "poscar"}, iterations=1)
        elif method == "zdd_count":
            benchmark.pedantic(zse.count, iterations=1)


@pytest.mark.parametrize("method", ["direct", "zdd", "zdd_count"])
@pytest.mark.benchmark(group="basic-ternary")
def test_basic_ternary(method, benchmark):
    setting = {
        "base_structure": get_lattice("fcc"),
        "index": 10,
        "num_types": 3,
        "remove_incomplete": True,
        "remove_superperiodic": True,
    }

    if method == "direct":
        se = StructureEnumerator(**setting)
        benchmark.pedantic(se.generate, kwargs={"output": "poscar"}, iterations=1)
    else:
        zse = ZddStructureEnumerator(**setting)
        if method == "zdd":
            benchmark.pedantic(zse.generate, kwargs={"output": "poscar"}, iterations=1)
        elif method == "zdd_count":
            benchmark.pedantic(zse.count, iterations=1)


def get_rutile_structure():
    # rutile structure taken from mp-856
    a = 4.832
    c = 3.243
    x_4f = 0.3066

    lattice = Lattice.from_parameters(a, a, c, 90, 90, 90)
    species = ["Sn", "Sn", "O", "O", "O", "O"]
    # fmt: off
    frac_coords = np.array([
        [0, 0, 0],                      # Sn(2a)
        [0.5, 0.5, 0.5],                # Sn(2a)
        [x_4f, x_4f, 0],                # O(4f)
        [1 - x_4f, 1 - x_4f, 0],        # O(4f)
        [0.5 - x_4f, 0.5 + x_4f, 0.5],  # O(4f)
        [0.5 + x_4f, 0.5 - x_4f, 0.5],  # O(4f)
    ])
    # fmt: on
    structure = Structure(lattice, species, frac_coords)
    return structure


@pytest.mark.parametrize("method", ["direct", "zdd", "zdd_count"])
@pytest.mark.benchmark(group="composition")
def test_composition_constraints(method, benchmark):
    setting = {
        "base_structure": get_rutile_structure(),
        "index": 2,
        "num_types": 3,
        "composition_constraints": [1, 3, 2],
        "remove_incomplete": True,
        "remove_superperiodic": True,
    }

    if method == "direct":
        se = StructureEnumerator(**setting)
        benchmark.pedantic(se.generate, kwargs={"output": "poscar"}, iterations=1)
    else:
        zse = ZddStructureEnumerator(**setting)
        if method == "zdd":
            benchmark.pedantic(zse.generate, kwargs={"output": "poscar"}, iterations=1)
        elif method == "zdd_count":
            benchmark.pedantic(zse.count, iterations=1)


@pytest.mark.parametrize("method", ["direct", "zdd", "zdd_count"])
@pytest.mark.benchmark(group="site")
def test_site_constraints(method, benchmark):
    setting = {
        "base_structure": get_rutile_structure(),
        "index": 4,
        "num_types": 3,
        "base_site_constraints": [
            [2],  # 2a
            [2],  # 2a
            [0, 1],  # 4f
            [0, 1],  # 4f
            [0, 1],  # 4f
            [0, 1],  # 4f
        ],
        "remove_incomplete": True,
        "remove_superperiodic": True,
    }

    if method == "direct":
        se = StructureEnumerator(**setting)
        benchmark.pedantic(se.generate, kwargs={"output": "poscar"}, iterations=1)
    else:
        zse = ZddStructureEnumerator(**setting)
        if method == "zdd":
            benchmark.pedantic(zse.generate, kwargs={"output": "poscar"}, iterations=1)
        elif method == "zdd_count":
            benchmark.pedantic(zse.count, iterations=1)


@pytest.mark.parametrize("method", ["direct", "zdd", "zdd_count"])
@pytest.mark.benchmark(group="composition-site")
def test_composition_site_constraints(method, benchmark):
    setting = {
        "base_structure": get_rutile_structure(),
        "index": 4,
        "num_types": 3,
        "composition_constraints": [1, 3, 2],
        "base_site_constraints": [
            [2],  # 2a
            [2],  # 2a
            [0, 1],  # 4f
            [0, 1],  # 4f
            [0, 1],  # 4f
            [0, 1],  # 4f
        ],
        "remove_incomplete": True,
        "remove_superperiodic": True,
    }

    if method == "direct":
        se = StructureEnumerator(**setting)
        benchmark.pedantic(se.generate, kwargs={"output": "poscar"}, iterations=1)
    else:
        zse = ZddStructureEnumerator(**setting)
        if method == "zdd":
            benchmark.pedantic(zse.generate, kwargs={"output": "poscar"}, iterations=1)
        elif method == "zdd_count":
            benchmark.pedantic(zse.count, iterations=1)



import pytest

from dsenum.superlattice import (
    generate_all_superlattices,
    generate_symmetry_distinct_superlattices,
)
from dsenum.utils import get_lattice


def test_generate_all_superlattices():
    # https://oeis.org/A001001
    num_expected = [
        1,
        7,
        13,
        35,
        31,
        91,
        57,
        155,
        130,
        217,
        133,
        455,
        183,
        399,
        403,
        651,
        307,
        910,
        381,
        1085,
        741,
        931,
        553,
        2015,
        806,
        1281,
        1210,
        1995,
        871,
        2821,
        993,
        2667,
        1729,
        2149,
        1767,
        4550,
        1407,
        2667,
        2379,
        4805,
        1723,
        5187,
        1893,
        4655,
        4030,
        3871,
        2257,
        8463,
        2850,
        5642,
        3991,
        6405,
        2863,
    ]
    max_index = len(num_expected)

    for index, expected in zip(range(1, max_index + 1), num_expected):
        list_HNF = generate_all_superlattices(index)
        assert len(list_HNF) == expected


@pytest.mark.parametrize(
    "kind,list_expected",
    [
        ("fcc", [1, 2, 3, 7, 5, 10, 7, 20, 14, 18]),
        ("bcc", [1, 2, 3, 7, 5, 10, 7, 20, 14, 18]),
        ("sc", [1, 3, 3, 9, 5, 13, 7, 24, 14, 23]),
        ("hex", [1, 3, 5, 11, 7, 19, 11, 34, 23, 33]),
        ("tet", [1, 5, 5, 17, 9, 29, 13, 51, 28, 53]),
        ("hcp", [1, 3, 5, 11, 7, 19, 11, 34, 23, 33]),
    ],
)
def test_reduce_HNF_list_by_parent_lattice_symmetry(kind, list_expected):
    # Confirm table 4
    structure = get_lattice(kind)
    for index, expected in zip(range(1, len(list_expected) + 1), list_expected):
        list_reduced_HNF = generate_symmetry_distinct_superlattices(index, structure)
        assert len(list_reduced_HNF) == expected


@pytest.mark.parametrize("kind", ["fcc", "bcc"])
def test_reduce_HNF_list_by_parent_lattice_symmetry_fcc_bcc(kind):
    # https://oeis.org/A045790
    lst_num = [
        1,
        2,
        3,
        7,
        5,
        10,
        7,
        20,
        14,
        18,
        11,
        41,
        15,
        28,
        31,
        58,
        21,
        60,
        25,
        77,
        49,
        54,
        33,
        144,
        50,
        72,
        75,
        123,
        49,
        158,
        55,
        177,
        97,
        112,
        99,
        268,
        75,
        136,
        129,
        286,
        89,
        268,
        97,
        249,
        218,
        190,
        113,
        496,
        146,
        280,
    ]

    structure = get_lattice(kind)
    for index, expected in zip(range(1, len(lst_num) + 1), lst_num):
        if index >= 20:
            continue
        list_reduced_HNF = generate_symmetry_distinct_superlattices(index, structure)
        assert len(list_reduced_HNF) == expected



import numpy as np
from pymatgen.analysis.structure_analyzer import SpacegroupAnalyzer
from pymatgen.analysis.structure_prediction.volume_predictor import DLSVolumePredictor
from pymatgen.core import Lattice, Structure
from pymatgen.core.periodic_table import DummySpecie
from pymatgen.io.cif import CifWriter

from dsenum._version import __version__


def get_lattice(kind):
    if kind == "hcp":
        latt = Lattice(np.array([[1, 0, 0], [0.5, np.sqrt(3) / 2, 0], [0, 0, 2 * np.sqrt(6) / 3]]))
        coords = [[0, 0, 0], [1 / 3, 1 / 3, 0.5]]
    else:
        coords = [[0, 0, 0]]

    if kind == "sc":
        latt = Lattice(np.eye(3))
    elif kind == "fcc":
        latt = Lattice(np.array([[0, 1, 1], [1, 0, 1], [1, 1, 0]]))
    elif kind == "bcc":
        latt = Lattice(np.array([[-1, 1, 1], [1, -1, 1], [1, 1, -1]]))
    elif kind == "hex":
        latt = Lattice.hexagonal(1, 2 * np.sqrt(6) / 3)
    elif kind == "tet":
        latt = Lattice(np.diag([1, 1, 1.2]))

    struct = Structure(latt, [DummySpecie("X")] * len(coords), coords)
    return struct


def square2d_lattice_symmetry():
    """
    symmetry operations of 2D square lattice (p4mm)
    """
    # p4mm
    # fmt: off
    rotations = np.array([
        [[1, 0],
         [0, 1]],   # x, y
        [[-1, 0],
         [0, -1]],  # -x, -y
        [[0, -1],
         [1, 0]],   # -y, x
        [[0, 1],
         [-1, 0]],  # y, -x
        [[-1, 0],
         [0, 1]],  # -x, y
        [[1, 0],
         [0, -1]],  # x, -y
        [[0, 1],
         [1, 0]],  # y, x
        [[0, -1],
         [-1, 0]],  # -y, -x
    ])
    translations = np.array([
        [0, 0],
        [0, 0],
        [0, 0],
        [0, 0],
        [0, 0],
        [0, 0],
        [0, 0],
        [0, 0],
    ])
    # fmt: on
    return rotations, translations


def get_symmetry_operations(structure, symprec: float = 1e-2):
    """
    find symmetry operations for a given structure

    Parameters
    ----------
    structure: pymatgen.core.Structure
    symprec: precision parameter in spglib

    Returns
    -------
    rotations: array, (# of symmetry operations, 3, 3)
    translations: array, (# of symmetry operations, 3)
    """
    sym_dataset = SpacegroupAnalyzer(structure, symprec=symprec).get_symmetry_dataset()
    rotations = sym_dataset["rotations"]
    translations = sym_dataset["translations"]
    return rotations, translations


def write_cif(
    filename: str,
    struct: Structure,
    refine_cell=False,
    resize_volume=False,
    symprec=1e-2,
    comment="",
):
    """
    dump structure in CIF format after resizing to feasible volume and refing cell by symmetry.

    Parameters
    ----------
    filename: str
    struct: pymatgen.core.Structure
    refine_cell: bool, optional
        if true, refine cell setting by spglib
    resize_volume: bool, optional
        if true, resize lattice by DLSVolumePredictor in pymatgen
    symprec: float, optional
        symprec in spglib
    """
    struct = refine_and_resize_structure(struct, refine_cell, resize_volume)
    cw = CifWriter(struct, symprec=symprec)
    comment = f"# generated by {__version__}\n" + comment
    with open(filename, "w") as f:
        if comment:
            f.write(comment + "\n" + cw.__str__())
        else:
            f.write(cw.__str__())


def refine_and_resize_structure(struct, refine_cell=True, resize_volume=True):
    if resize_volume:
        dls = DLSVolumePredictor()
        struct = dls.get_predicted_structure(struct)
        struct.apply_strain(0.5)

    if refine_cell:
        sga = SpacegroupAnalyzer(struct)
        # SpacegroupAnalyzer.get_primitive_standard_structure may return incorrect structures
        # So we avoid to use SpacegroupAnalyzer.get_primitive_standard_structure
        struct = sga.get_refined_structure()

    return struct


def cast_integer_matrix(arr: np.ndarray) -> np.ndarray:
    arr_int = np.around(arr).astype(int)
    return arr_int



from itertools import product
from typing import List, Optional, Tuple, Union, cast

import numpy as np
from hsnf import smith_normal_form

from dsenum.site import CanonicalSite, DerivativeSite
from dsenum.utils import cast_integer_matrix

"""
                hash_frac_coords               ravel_canonical_site
    frac_coords ---------------> CanonicalSite -------------------------> Int
       |                            |^         <-------------------------
       |                            ||          unravel_to_canonical_site
       |                            ||
       |    embed_to_derivative_site||hash_derivative_site
       |                            ||
       |                            V|
       |-----------------------> DerivativeSite
            get_frac_coords
"""


class DerivativeMultiLatticeHash:
    """
    Parameters
    ----------
    hnf: array, (dim, dim)
        Hermite normal form(lower triangular)
    displacement_set: array, (num_site_parent, dim)
        fractinal coordinates in primitive cell of base structure
    """

    def __init__(self, hnf: np.ndarray, displacement_set: np.ndarray):
        self._hnf = hnf
        self._index = np.around(np.abs(np.linalg.det(self.hnf))).astype(int)
        assert self.index != 0

        self.displacement_set = displacement_set
        assert self.dim == self.displacement_set.shape[1]

        self.num_site_base = len(displacement_set)

        D, L, R = smith_normal_form(self.hnf)
        self.snf = D
        self.left = L
        self.right = R
        self.left_inv = cast_integer_matrix(np.linalg.inv(self.left))

        self.invariant_factors = tuple(self.snf.diagonal())
        self.shape = (self.num_site_base,) + self.invariant_factors

    @property
    def hnf(self) -> np.ndarray:
        return self._hnf

    @property
    def dim(self) -> int:
        return self.hnf.shape[0]

    @property
    def index(self) -> int:
        return self._index

    @property
    def num_sites(self) -> int:
        return self.num_site_base * self.index

    @property
    def num_base_sites(self) -> int:
        return self.num_site_base

    def hash_derivative_site(
        self, dsite: DerivativeSite, return_image: bool = False
    ) -> Union[CanonicalSite, Tuple[CanonicalSite, np.ndarray]]:
        """
        Returns
        -------
        csite: CanoncalSite
        derivative_jimage: (Optional), array

        self.get_frac_coords(dsite) == self.displacement_set[dsite.site_index]
                                        + np.dot(self.left_inv, csite.factor)
                                        + np.dot(self.hnf, derivative_jimage)
        """
        site_index, jimage = dsite.site_index, dsite.jimage

        factor_tmp = np.dot(self.left, np.array(jimage, dtype=int))
        factor = np.mod(factor_tmp, np.array(self.invariant_factors))

        csite = CanonicalSite(site_index, tuple(factor))

        if return_image:
            derivative_jimage = cast_integer_matrix(factor_tmp - factor) / np.array(
                self.invariant_factors
            )
            derivative_jimage = cast_integer_matrix(derivative_jimage)
            derivative_jimage = np.dot(self.right, derivative_jimage).astype(int)
            return csite, derivative_jimage
        else:
            return csite

    def hash_frac_coords(self, frac_coord: np.ndarray) -> Optional[CanonicalSite]:
        for site_index, fc in enumerate(self.displacement_set):
            jimage = cast_integer_matrix(frac_coord - fc)
            if np.allclose(fc + jimage, frac_coord):
                dsite = DerivativeSite(site_index, jimage)
                csite = cast(CanonicalSite, self.hash_derivative_site(dsite))
                return csite
        return None

    def get_canonical_sites_list(self) -> List[CanonicalSite]:
        list_csites = []
        for site_index in range(len(self.displacement_set)):
            for factor in product(*[range(f) for f in self.invariant_factors]):
                csite = CanonicalSite(site_index, tuple(factor))
                list_csites.append(csite)

        assert all([self.ravel_canonical_site(csite) == i for i, csite in enumerate(list_csites)])
        return list_csites

    def get_distinct_derivative_sites_list(self) -> List[DerivativeSite]:
        list_csites = self.get_canonical_sites_list()
        list_dsites = [self.embed_to_derivative_site(csite) for csite in list_csites]
        return list_dsites

    def get_canonical_and_derivative_sites_list(
        self,
    ) -> List[Tuple[CanonicalSite, DerivativeSite]]:
        list_csites_dsites = []
        for site_index in range(len(self.displacement_set)):
            for factor in product(*[range(f) for f in self.invariant_factors]):
                csite = CanonicalSite(site_index, tuple(factor))
                dsite = self.embed_to_derivative_site(csite)
                list_csites_dsites.append((csite, dsite))

        return list_csites_dsites

    def get_lattice_points(self) -> List[np.ndarray]:
        """
        return list of lattice points with the supercell
        Note that the returned lattice points are not guaranteed to be "within" the supercell
        """
        lattice_points = []
        for factor in self.get_all_factors():
            lp = cast_integer_matrix(np.dot(self.left_inv, factor))
            lattice_points.append(lp)
        return lattice_points

    def get_all_factors(self) -> List[Tuple[int, ...]]:
        list_factors = list(product(*[range(f) for f in self.invariant_factors]))
        return list_factors

    def modulus_factor(self, factor: np.ndarray) -> np.ndarray:
        modded_factor = np.mod(factor, np.array(self.invariant_factors))
        return modded_factor

    def get_frac_coords(self, dsite: DerivativeSite) -> np.ndarray:
        return self.displacement_set[dsite.site_index] + np.array(dsite.jimage)

    def ravel_canonical_site(self, csite: CanonicalSite) -> int:
        # hash canonical site s.t. self.get_canonical_sites_list <=> identity
        multi_index = (csite.site_index,) + tuple(csite.factor)
        raveled: int = np.ravel_multi_index(multi_index, self.shape)
        return raveled

    def ravel_derivative_site(self, dsite: DerivativeSite) -> int:
        csite = self.hash_derivative_site(dsite)
        raveled: int = self.ravel_canonical_site(csite)  # type: ignore
        return raveled

    def unravel_to_canonical_site(self, indices: int) -> CanonicalSite:
        unraveled = np.unravel_index(indices, self.shape)
        site_index, factor = unraveled[0], unraveled[1:]
        csite = CanonicalSite(site_index, factor)
        return csite

    def embed_to_derivative_site(self, csite: CanonicalSite) -> DerivativeSite:
        jimage_base = cast_integer_matrix(np.dot(self.left_inv, csite.factor))
        dsite = DerivativeSite(csite.site_index, tuple(jimage_base.tolist()))
        return dsite


def get_species_list(index: int, list_species: List[str]) -> List[str]:
    """
    tile list of species for derivative structure
    """
    species = []
    for sp in list_species:
        species.extend([sp] * index)
    return species


def convert_site_constraints(
    base_site_constraints: List[List[int]], index: int
) -> List[List[int]]:
    site_constraints = []
    for sc in base_site_constraints:
        for _ in range(index):
            site_constraints.append(sc)
    return site_constraints



from abc import ABCMeta, abstractmethod
from itertools import product
from typing import List

import numpy as np
from sympy.utilities.iterables import multiset_permutations

from dsenum.core import get_composition, hash_in_all_configuration  # type: ignore


class BaseColoringGenerator(metaclass=ABCMeta):
    @abstractmethod
    def generate_all_colorings(self):
        raise NotImplementedError

    @abstractmethod
    def yield_coloring(self):
        raise NotImplementedError


class ColoringGenerator(BaseColoringGenerator):
    def __init__(self, num_elements: int, num_color: int, site_constraints=None):
        """
        Parameters
        ----------
        num_elements:
            the number of elements to color
        num_color:
            the number of color
        site_constraints: list of list of int, optional
            the length of this list should be equal to `num_elements`
        """
        self.num_elements = num_elements
        self.num_color = num_color

        if site_constraints is not None:
            assert len(site_constraints) == self.num_elements
        self.site_constraints = site_constraints

    def generate_all_colorings(self):
        if self.site_constraints:
            list_colorings = []
            flags = dict()
            for cl_compressed in product(*[range(len(sc)) for sc in self.site_constraints]):
                cl = [self.site_constraints[i][idx] for i, idx in enumerate(cl_compressed)]
                list_colorings.append(cl)
                flags[hash_in_all_configuration(cl, self.num_color)] = True
        else:
            list_colorings = list(product(range(self.num_color), repeat=self.num_elements))
            flags = {
                hash_in_all_configuration(list(coloring), self.num_color): True
                for coloring in list_colorings
            }

        return list_colorings, flags

    def yield_coloring(self):
        if self.site_constraints:
            for cl_compressed in product(*[range(len(sc)) for sc in self.site_constraints]):
                cl = [self.site_constraints[i][idx] for i, idx in enumerate(cl_compressed)]
                yield cl
        else:
            for cl in product(range(self.num_color), repeat=self.num_elements):
                yield list(cl)


class ListBasedColoringGenerator(BaseColoringGenerator):
    """
    Parameters
    ----------
    num_color: int
    list_colorings: list of generator
    """

    def __init__(self, num_color, list_colorings):
        self.num_color = num_color
        self.list_colorings = list_colorings

    def generate_all_colorings(self):
        flags = {
            hash_in_all_configuration(coloring, self.num_color): True
            for coloring in self.list_colorings
        }
        return self.list_colorings, flags

    def yield_coloring(self):
        yield from self.list_colorings


class FixedConcentrationColoringGenerator(BaseColoringGenerator):
    """
    Parameters
    ----------
    num_elements: int
    num_color: int
    color_ratio: list of int
        length of `color_ratio` should be equal to `num_color`.
        num_elements % sum(color_ratio) == 0
    site_constraints: (Optional), list (num_elements, num_color)
        e.g. site_constraints[2] = [0, 3, 4] means color of site-2 must be 0, 3, or 4.
    """

    def __init__(
        self, num_elements: int, num_color: int, color_ratio: List[float], site_constraints=None
    ):
        self.num_elements = num_elements
        self.num_color = num_color

        self.site_constraints = site_constraints

        self.color_ratio = color_ratio
        ratio_sum = int(np.around(sum(self.color_ratio)))
        if num_elements % ratio_sum != 0:
            raise ValueError("incorrect composition ratio")

        factor = num_elements // ratio_sum
        self.num_elements_each_color = [int(np.around(factor * cr)) for cr in self.color_ratio]

    def generate_all_colorings(self):
        if self.site_constraints:
            list_colorings = []
            flags = dict()
            # Apply site_constraints first
            for cl_compressed in product(*[range(len(sc)) for sc in self.site_constraints]):
                cl = [self.site_constraints[i][idx] for i, idx in enumerate(cl_compressed)]
                if get_composition(cl, self.num_color) == self.num_elements_each_color:
                    list_colorings.append(cl)
                    flags[hash_in_all_configuration(cl, self.num_color)] = True
        else:
            first_coloring = []
            for i in range(self.num_color):
                first_coloring.extend([i for _ in range(self.num_elements_each_color[i])])

            # TODO: inefficient to cast to list
            list_colorings = list(multiset_permutations(first_coloring))
            flags = {
                hash_in_all_configuration(coloring, self.num_color): True
                for coloring in list_colorings
            }
        return list_colorings, flags

    def yield_coloring(self):
        # create one of colorings to use multiset_permutations
        first_coloring = []
        for i in range(self.num_color):
            first_coloring.extend([i for _ in range(self.num_elements_each_color[i])])

        if self.site_constraints:
            for cl in multiset_permutations(first_coloring):
                # TODO: more efficient way to adopt site_constraints
                if satisfy_site_constraints(self.site_constraints, cl):
                    yield cl
        else:
            for cl in multiset_permutations(first_coloring):
                yield cl


def satisfy_site_constraints(site_constraints, coloring):
    if all([(color in site_constraints[i]) for i, color in enumerate(coloring)]):
        return True
    else:
        return False



import numpy as np

from dsenum.utils import cast_integer_matrix


def test_cast_integer_matrix():
    arr = np.array([1.0, 2.0001, 2.9999, -0.9999, -2.0001])
    arr_rounded = np.array([1.0, 2.0, 3.0, -1.0, -2.0])
    assert np.allclose(cast_integer_matrix(arr), arr_rounded)



from abc import ABCMeta, abstractmethod
from itertools import permutations
from multiprocessing import Pool, cpu_count
from typing import List, Tuple, cast

from dsenum.coloring_generator import BaseColoringGenerator
from dsenum.core import act_permutation, hash_in_all_configuration  # type: ignore
from dsenum.permutation_group import DerivativeStructurePermutation


class AbstractEnumerator(metaclass=ABCMeta):
    @abstractmethod
    def coset_enumerate(self) -> List[List[int]]:
        raise NotImplementedError


class DirectColoringEnumerator(AbstractEnumerator):
    """
    Parameters
    ----------
    permutation_group: list of permutation
    num_color: int
    cl_generator: BaseColoringGenerator
    color_exchange: bool
    """

    def __init__(
        self,
        permutation_group: List[List[int]],
        num_color: int,
        cl_generator: BaseColoringGenerator,
        color_exchange: bool = True,
    ):
        self.permutation_group = permutation_group
        self.num_color = num_color
        self.cl_generator = cl_generator
        self.color_exchange = color_exchange

        self.list_colorings, self.flags = self.cl_generator.generate_all_colorings()

    def _hash(self, coloring: List[int]) -> int:
        return hash_in_all_configuration(coloring, self.num_color)

    def _walk_orbit(self, coloring: List[int], include_identity: bool = False) -> None:
        offset = 0 if include_identity else 1
        # assume self.permutation_group[0] is identity
        for prm in self.permutation_group[offset:]:
            acted_cl = act_permutation(prm, coloring)
            acted_cl_hash = self._hash(acted_cl)
            self.flags[acted_cl_hash] = False

    def coset_enumerate(self) -> List[List[int]]:
        colorings = []

        for cl in self.list_colorings:
            cl_hash = self._hash(cl)
            # avoid already-visited coloring
            if not self.flags[cl_hash]:
                continue
            colorings.append(cl)

            self._walk_orbit(cl, include_identity=False)

            if self.color_exchange:
                for cl_prm in permutations(range(self.num_color)):
                    exchanged_cl = [cl_prm[c] for c in cl]
                    exchanged_cl_hash = self._hash(exchanged_cl)
                    if exchanged_cl_hash == cl_hash:
                        continue
                    self._walk_orbit(exchanged_cl, include_identity=True)

        return colorings


class LexicographicColoringEnumerator(AbstractEnumerator):
    """
    this algorithm takes the most lexicographically small colorings

    Parameters
    ----------
    permutation_group: list of permutation
    num_color: int
    cl_generator: BaseColoringGenerator
    color_exchange: bool
    n_jobs: int
        when n_jobs > 1, use multiprocessing but it seems slower than signle core......
    """

    def __init__(
        self,
        permutation_group: List[List[int]],
        num_color: int,
        cl_generator: BaseColoringGenerator,
        color_exchange: bool = True,
        n_jobs: int = 1,
    ):
        self.permutation_group = permutation_group
        self.num_color = num_color
        self.cl_generator = cl_generator
        self.color_exchange = color_exchange
        if n_jobs == -1:
            self.n_jobs = cpu_count()
        else:
            self.n_jobs = n_jobs

    def _hash(self, coloring: List[int]) -> int:
        return hash_in_all_configuration(coloring, self.num_color)

    def _is_champion_coloring(self, coloring: List[int]) -> bool:
        cl_hash = self._hash(coloring)

        if self.color_exchange:
            for cl_prm in permutations(range(self.num_color)):
                exchanged_cl = [cl_prm[c] for c in coloring]
                for prm in self.permutation_group:
                    acted_cl = act_permutation(prm, exchanged_cl)
                    acted_cl_hash = self._hash(acted_cl)
                    if acted_cl_hash < cl_hash:
                        return False
        else:
            for prm in self.permutation_group:
                acted_cl = act_permutation(prm, coloring)
                acted_cl_hash = self._hash(acted_cl)
                if acted_cl_hash < cl_hash:
                    return False
        return True

    def _is_champion_coloring_parallel(self, coloring: List[int]) -> Tuple[List[int], bool]:
        return coloring, self._is_champion_coloring(coloring)

    def coset_enumerate(self) -> List[List[int]]:
        if self.n_jobs != 1:
            with Pool(self.n_jobs) as pool:
                colorings = [
                    cl
                    for cl, keep in pool.map(
                        self._is_champion_coloring_parallel, self.cl_generator.yield_coloring()
                    )
                    if keep
                ]
        else:
            colorings = []
            for cl in self.cl_generator.yield_coloring():
                if self._is_champion_coloring(cl):
                    colorings.append(cl)
        return colorings


class SiteColoringEnumerator:
    """
    Parameters
    ----------
    ds_permutation: DerivativeStructurePermutation object
    num_color: int
    color_exchange: bool
    remove_superperiodic: bool
    method: "direct" or "lexicographic", so far
    n_jobs: int, use only when method is "lexicographic"
    """

    def __init__(
        self,
        num_color: int,
        ds_permutation: DerivativeStructurePermutation,
        cl_generator: BaseColoringGenerator,
        color_exchange: bool = True,
        remove_superperiodic: bool = True,
        remove_incomplete: bool = True,
        method: str = "direct",
        n_jobs: int = 1,
    ):
        self.num_color = num_color
        self.ds_permutation = ds_permutation
        self.cl_generator = cl_generator
        self.color_exchange = color_exchange
        self.remove_superperiodic = remove_superperiodic
        self.remove_incomplete = remove_incomplete
        self.method = method
        self.n_jobs = n_jobs

        self.permutation_group = self.ds_permutation.get_symmetry_operation_permutations()

        # typing.cast causes no runtime effect
        if self.method == "direct":
            self.clenum = cast(
                AbstractEnumerator,
                DirectColoringEnumerator(
                    self.permutation_group,
                    self.num_color,
                    self.cl_generator,
                    color_exchange=color_exchange,
                ),
            )
        elif self.method == "lexicographic":
            self.clenum = cast(
                AbstractEnumerator,
                LexicographicColoringEnumerator(
                    self.permutation_group,
                    self.num_color,
                    self.cl_generator,
                    color_exchange=color_exchange,
                    n_jobs=self.n_jobs,
                ),
            )
        else:
            raise ValueError("Unknown method: ", self.method)

    @property
    def translation_permutations(self) -> List[List[int]]:
        return self.ds_permutation.prm_t

    def _hash(self, coloring: List[int]) -> int:
        return hash_in_all_configuration(coloring, self.num_color)

    def unique_colorings(self) -> List[List[int]]:
        symmetry_uniqued_coloring = self.clenum.coset_enumerate()
        colorings = []

        for cl in symmetry_uniqued_coloring:
            if self.remove_incomplete and (not self._has_all_colors(cl)):
                continue
            if self.remove_superperiodic and self._is_superperiodic(cl):
                continue
            colorings.append(cl)

        return colorings

    def _has_all_colors(self, coloring: List[int]) -> bool:
        if len(set(coloring)) == self.num_color:
            return True
        else:
            return False

    def _is_superperiodic(self, coloring: List[int]) -> bool:
        # assume self.translation_permutations[0] is identity
        if any(
            [
                self._hash(act_permutation(prm, coloring)) == self._hash(coloring)
                for prm in self.translation_permutations[1:]
            ]
        ):
            return True
        else:
            return False



from functools import lru_cache
from itertools import product
from typing import List

from scipy.special import binom


def polya_counting(permutation_group: List[List[int]], num_color: int) -> int:
    """
    count the number of coloring with num_color kinds of colors and permutations

    Parameters
    ----------
    permutation_group:
        the i-th permutation permutation_group[i] permutes j to permutation_group[i][j]
        for j = 0,...,len(permutation_group[0])-1

    Returns
    -------
    cnt: int
    """
    cnt = 0
    for perm in permutation_group:
        type_of_perm = get_type_of_permutation(perm)
        cnt += num_color ** sum(type_of_perm)

    assert cnt % len(permutation_group) == 0
    cnt //= len(permutation_group)
    return cnt


@lru_cache(maxsize=None)
def get_inventory_coefficient(type_of_perm: tuple, didx, num_elements_of_each_color: tuple):
    # termination
    if didx == 0:
        if all([e == 0 for e in num_elements_of_each_color]):
            return 1
        else:
            return 0

    ret = 0

    for nec in product(*[range(0, e + 1, didx) for e in num_elements_of_each_color]):
        list_k = [e // didx for e in nec]
        if sum(list_k) != type_of_perm[didx - 1]:
            continue
        complement = tuple(e1 - e2 for e1, e2 in zip(num_elements_of_each_color, nec))
        ret += get_multinomial_coefficient(list_k) * get_inventory_coefficient(
            type_of_perm, didx - 1, complement
        )

    return ret


# noqa: B950 https://stackoverflow.com/questions/46374185/does-python-have-a-function-which-computes-multinomial-coefficients
def get_multinomial_coefficient(params):
    if len(params) == 1:
        return 1
    if params[-1] == 0:
        return get_multinomial_coefficient(params[:-1])
    return int(binom(sum(params), params[-1])) * get_multinomial_coefficient(params[:-1])


def polya_fixed_degrees_counting(
    permutation_group: List[List[int]], num_color: int, num_elements_of_each_color: List[int]
):
    """
    count the number of coloring with num_color kinds of colors and permutations.
    In addition, the number of each colors is fixed with num_elements_of_each_color.

    Parameters
    ----------
    permutation_group:
        the i-th permutation permutation_group[i] permutes j to permutation_group[i][j]
        for j = 0,...,len(permutation_group[0])-1
    num_color: int
    num_elements_of_each_color: List of int, (num_color)

    Returns
    -------
    coeffs: int
    """
    num_elements = len(permutation_group[0])
    coeffs = 0

    for perm in permutation_group:
        type_of_perm = tuple(get_type_of_permutation(perm))
        coeffs_perm = get_inventory_coefficient(
            type_of_perm, num_elements, tuple(num_elements_of_each_color)
        )
        coeffs += coeffs_perm

    assert coeffs % len(permutation_group) == 0
    coeffs //= len(permutation_group)
    return coeffs


def get_type_of_permutation(permutation: List[int]):
    num_elements = len(permutation)
    type_of_perm = [0 for _ in range(num_elements)]

    flags = [False for _ in range(num_elements)]
    for i in range(num_elements):
        if flags[i]:
            continue
        flags[i] = True
        pos = permutation[i]
        cnt = 1
        while pos != i:
            flags[pos] = True
            pos = permutation[pos]
            cnt += 1

        type_of_perm[cnt - 1] += 1

    return type_of_perm



import numpy as np
import pytest
from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.core import Lattice, Structure
from pymatgen.core.periodic_table import DummySpecie, Element
from pymatgen.io.vasp.inputs import Poscar

from dsenum import StructureEnumerator
from dsenum.utils import get_lattice


@pytest.mark.parametrize("kind", ["fcc", "sc", "hcp"])
@pytest.mark.parametrize("index", [2, 3, 4])
def test_basic_structure(kind, index):
    base_structure = get_lattice(kind)
    num_type = 2
    species = ["Cu", "Au"]

    se = StructureEnumerator(
        base_structure,
        index,
        num_type,
        species,
        color_exchange=True,
        remove_superperiodic=True,
    )
    list_ds = se.generate()

    stm = StructureMatcher(ltol=1e-4, stol=1e-4)
    grouped = stm.group_structures(list_ds)
    assert len(grouped) == len(list_ds)


@pytest.fixture
def sto_perovskite() -> Structure:
    lattice = Lattice(3.945 * np.eye(3))
    species = ["Sr", "Ti", "O", "O", "O"]
    frac_coords = np.array(
        [[0, 0, 0], [0.5, 0.5, 0.5], [0.0, 0.5, 0.5], [0.5, 0.0, 0.5], [0.5, 0.5, 0.0]]
    )
    return Structure(lattice, species, frac_coords)


def test_coloring_with_fixed_species(sto_perovskite: Structure):
    base_structure = sto_perovskite.copy()
    base_structure.remove_species(["Sr", "Ti"])
    additional_species = sto_perovskite.species[:2]
    additional_frac_coords = sto_perovskite.frac_coords[:2]

    mapping_color_species = [DummySpecie("X"), "O"]
    num_types = len(mapping_color_species)
    index = 2

    se = StructureEnumerator(
        base_structure,
        index,
        num_types,
        mapping_color_species=mapping_color_species,
        color_exchange=False,
        remove_superperiodic=True,
        remove_incomplete=False,
    )
    list_dstructs = se.generate(
        additional_species=additional_species, additional_frac_coords=additional_frac_coords
    )

    # with base_site_constraints
    mapping_color_species2 = [DummySpecie("X"), "O", "Sr", "Ti"]
    num_types2 = len(mapping_color_species2)
    base_site_constraints = [
        [2],  # Sr site
        [3],  # Cu site
        [0, 1],  # O or V
        [0, 1],  # O or V
        [0, 1],  # O or V
    ]
    se2 = StructureEnumerator(
        sto_perovskite,
        index,
        num_types2,
        mapping_color_species=mapping_color_species2,
        base_site_constraints=base_site_constraints,
        color_exchange=False,
        remove_superperiodic=True,
        remove_incomplete=False,
    )
    list_dstructs2 = se2.generate()

    # check uniqueness by StructureMatcher
    stm = StructureMatcher(ltol=1e-4, stol=1e-4)
    grouped = stm.group_structures(list_dstructs + list_dstructs2)  # type: ignore
    assert len(grouped) == len(list_dstructs)
    assert all([(len(matched) == 2) for matched in grouped])


@pytest.mark.parametrize(
    "kind",
    [
        "sc",
    ],
)
def test_poscar_string(kind):
    base_structure = get_lattice(kind)
    num_type = 2
    index = 4
    species = [Element("Cu"), Element("Au")]

    se = StructureEnumerator(
        base_structure,
        index,
        num_type,
        species,
        color_exchange=True,
        remove_superperiodic=True,
    )
    list_ds_mg = se.generate(output="pymatgen")
    list_ds_pc = se.generate(output="poscar")

    assert len(list_ds_mg) == len(list_ds_pc)
    for expect, poscar_str in zip(list_ds_mg, list_ds_pc):
        actual = Poscar.from_str(poscar_str).structure
        assert expect == actual



from hsnf import smith_normal_form

from dsenum.superlattice import generate_all_superlattices


def test_number_of_snf():
    # confirm table-3
    num_hnf_expected = [1, 7, 13, 35, 31, 91, 57, 155, 130, 217, 133, 455, 183, 399, 403, 651]
    num_snf_expected = [1, 1, 1, 2, 1, 1, 1, 3, 2, 1, 1, 2, 1, 1, 1, 4]
    max_index = len(num_hnf_expected)

    for index, hnf_expected, snf_expected in zip(
        range(1, max_index + 1), num_hnf_expected, num_snf_expected
    ):
        list_HNF = generate_all_superlattices(index)
        assert len(list_HNF) == hnf_expected

        list_SNF = set()
        for hnf in list_HNF:
            snf, _, _ = smith_normal_form(hnf)
            dag = tuple(snf.diagonal())
            list_SNF.add(dag)

        assert len(list_SNF) == snf_expected



cimport cython


@cython.boundscheck(False)
@cython.wraparound(False)
def hash_in_all_configuration(coloring, num_color):
    ret = 0
    cdef int e
    for e in coloring:
        ret = ret * num_color + e
    return ret


@cython.boundscheck(False)
@cython.wraparound(False)
def act_permutation(perm, coloring):
    new_coloring = [coloring[perm[i]] for i in range(len(coloring))]
    return new_coloring


@cython.boundscheck(False)
@cython.wraparound(False)
def get_composition(coloring, num_color):
    composition = [0 for _ in range(num_color)]
    for c in coloring:
        composition[c] += 1
    return composition



from copy import deepcopy
from dataclasses import astuple, dataclass
from itertools import combinations
from typing import List, Tuple

import numpy as np
from pymatgen.core import Structure

from dsenum.converter import DerivativeMultiLatticeHash
from dsenum.site import DerivativeSite
from dsenum.utils import cast_integer_matrix, get_symmetry_operations


@dataclass
class PointCluster:
    """
    cluster class represents the set of DerivativeSite
    Attributes
    ----------
    points: list of Point
    """

    points: List[DerivativeSite]

    def __str__(self):
        points_str = [f"site_index={p.site_index}, jimage={p.jimage}" for p in self.points]
        rep = "PointCluster[" + ";".join(points_str) + "]"
        return rep

    def __eq__(self, other):
        if not isinstance(other, PointCluster):
            return False

        return tuple(self.points) == tuple(other.points)

    def __hash__(self):
        return hash(tuple(self.points))

    def __len__(self):
        return len(self.points)


class EquivalentPointClusterGenerator:
    """
    find equivalent point clusters
    Parameters
    ----------
    frac_coords: np.ndarray, (None, dim)
    rotations: np.ndarray(None, dim, dim)
    translations: np.ndarray(None, dim, dim)
    converter: DerivativeMultiLatticeHash
        converter for periodic boundary condition
    """

    def __init__(
        self,
        frac_coords: np.ndarray,
        rotations: np.ndarray,
        translations: np.ndarray,
        converter: DerivativeMultiLatticeHash,
    ):
        self.frac_coords = frac_coords
        self.rotations = rotations
        self.translations = translations
        self.converter = converter

    @property
    def dim(self):
        return self.frac_coords.shape[1]

    def _get_dsite(self, frac_coord: np.ndarray) -> DerivativeSite:
        """
        return DerivativeSite from fractional coorinates

        Parameters
        ----------
        frac_coords: (dim, )

        Returns
        -------
        dsite: DerivativeSite
        """
        for site_index, fc in enumerate(self.frac_coords):
            jimage = cast_integer_matrix(frac_coord - fc)
            if np.allclose(fc + jimage, frac_coord):
                dsite = DerivativeSite(site_index, tuple(jimage.tolist()))
                return dsite
        raise ValueError(f"invalid fractional coordinates: {frac_coord}")

    def normalize_point_cluster(self, point_cluster: PointCluster) -> PointCluster:
        """
        normalize point-cluster such that,
            - jimage of cluster[0] is (0, 0, 0)
            - points are in ascending order
        """
        # sort sites by (site_index, jimage)
        tuple_points = sorted(astuple(p) for p in point_cluster.points)
        points = [DerivativeSite(site_index, jimage) for site_index, jimage in tuple_points]

        offset = np.array(points[0].jimage)
        shifed_new_points = [
            DerivativeSite(p.site_index, tuple((np.array(p.jimage, dtype=int) - offset).tolist()))
            for p in points
        ]
        return PointCluster(shifed_new_points)

    def operate_point_cluster(
        self, point_cluster: PointCluster, R: np.ndarray, tau: np.ndarray
    ) -> PointCluster:
        """
        operate (R, tau) to a point cluster
        """
        new_points = []
        for point in point_cluster.points:
            site_index, jimage = astuple(point)
            new_frac_coord = np.dot(R, self.frac_coords[site_index] + np.array(jimage)) + tau
            new_point = self._get_dsite(new_frac_coord)
            new_points.append(new_point)
        return PointCluster(new_points)

    def find_equivalent_point_clusters(self, point_cluster: PointCluster) -> List[PointCluster]:
        # (rotations, translations) should contain identity operation
        equiv_clusters = {
            self.normalize_point_cluster(self.operate_point_cluster(point_cluster, R, tau))
            for R, tau in zip(self.rotations, self.translations)
        }

        all_equiv_clusters = []
        # apply translation
        for cluster in equiv_clusters:
            for lp in self.converter.get_lattice_points():
                translated_cluster = self.operate_point_cluster(
                    cluster, np.eye(self.dim), np.array(lp)
                )
                all_equiv_clusters.append(translated_cluster)
        # to account for multiplicity, do not unique all_equiv_clusters
        return all_equiv_clusters

    def get_point_cluster_size(
        self, point_cluster: PointCluster, lattice_matrix: np.ndarray
    ) -> float:
        """
        cluster size is defined as a maximum distance between sites in a point cluster

        Parameters
        ----------
        lattice_matrix: (dim, dim)
            row-wise basis vectors
        """
        size = 0
        for s0, s1 in combinations(point_cluster.points, r=2):
            f0 = self.frac_coords[s0.site_index] + np.array(s0.jimage)
            f1 = self.frac_coords[s1.site_index] + np.array(s1.jimage)
            dist = np.linalg.norm(np.dot(lattice_matrix.T, f1 - f0))
            size = max(size, dist)
        return size

    def get_distinct_point_clusters(
        self, structure: Structure, cutoff: float, order=2, eps=1e-8
    ) -> List[PointCluster]:
        """
        return all distinct `order`-sites point-clusters

        Parameters
        ----------
        structure: Structure
        cutoff: float
            cutoff radius
        order: int

        Returns
        -------
        distinct_point_clusters: List[PointCluster]
            returned list of point clusters are sorted by cluster size in the ascending order.
        """
        # search neighbor points whose distnace from some site in the unit cell is less than cutoff.
        list_points: List[DerivativeSite] = []
        for site in structure:
            neighbors = structure.get_neighbors(site, cutoff)
            for s in neighbors:
                if s.nn_distance - cutoff > eps:
                    continue
                point = self._get_dsite(s.frac_coords)
                list_points.append(point)
        list_points = list(set(list_points))

        # Find distinct singlet clusters
        distinct_point_clusters: List[PointCluster] = []
        found_singlets = set()
        for site_index in range(len(self.frac_coords)):
            cluster = PointCluster([DerivativeSite(site_index, (0, 0, 0))])
            if cluster in found_singlets:
                continue
            distinct_point_clusters.append(cluster)
            for sym_cluster in self.find_equivalent_point_clusters(cluster):
                found_singlets.add(sym_cluster)

        # grow pairs to triplet, quadruple, ...
        for num_nn in range(2, order + 1):
            next_distinct_point_clusters: List[PointCluster] = []
            found = set()

            for added_point in list_points:
                for precluster in distinct_point_clusters:
                    points: List[DerivativeSite] = list(precluster.points) + [added_point]
                    point_cluster: PointCluster = self.normalize_point_cluster(
                        PointCluster(points)
                    )
                    # skip an overlapped cluster
                    if len(set(point_cluster.points)) != num_nn:
                        continue

                    # skip too large point cluster
                    cluster_size = self.get_point_cluster_size(
                        point_cluster, structure.lattice.matrix
                    )
                    if cluster_size - cutoff > eps:
                        continue

                    if point_cluster in found:
                        continue
                    next_distinct_point_clusters.append(point_cluster)

                    equiv_clusters = self.find_equivalent_point_clusters(point_cluster)
                    found.update(equiv_clusters)

            distinct_point_clusters = deepcopy(next_distinct_point_clusters)

        # sort by cluser size
        distinct_point_clusters.sort(
            key=lambda pcl: self.get_point_cluster_size(pcl, structure.lattice.matrix)
        )

        return distinct_point_clusters

    def get_all_clusters(
        self, structure: Structure, cutoff: float, order: int = 2, eps: float = 1e-8
    ) -> Tuple[List[List[PointCluster]], List[float]]:
        """
        return all `order`-sites clusters

        Parameters
        ----------
        structure: Structure
        cutoff: float
            cutoff radius
        order: int

        Returns
        -------
        grouped_point_clusters: List[List[PointCluster]]
            grouped_points_clusters[i] is a list of the i-th equivalent group of clusters
        point_cluster_sizes: List[float]
            point_cluster_sizes[i] is a size of the i-th equivalent group of clusters
        """
        # find distinct point clusters
        distinct_temp_point_clusters = self.get_distinct_point_clusters(
            structure, cutoff, order, eps
        )

        # generate symmetry-equivalent point clusters
        grouped_clusters_sizes = []
        for point_cluster in distinct_temp_point_clusters:
            cluster_size = self.get_point_cluster_size(point_cluster, structure.lattice.matrix)
            equivs = self.find_equivalent_point_clusters(point_cluster)
            grouped_clusters_sizes.append((equivs, cluster_size))

        # sort by cluster size
        grouped_clusters_sizes.sort(key=lambda e: e[1])

        grouped_point_clusters = [gcls for gcls, _ in grouped_clusters_sizes]
        point_cluster_sizes = [size for _, size in grouped_clusters_sizes]

        return grouped_point_clusters, point_cluster_sizes

    @classmethod
    def from_structure(
        cls, structure: Structure, transformation: np.ndarray, symprec: float = 1e-2
    ):
        """
        create EquivalentPointClusterGenerator from pymatgen.core.Structure

        Parameters
        ----------
        structure: pymatgen.core.Structure, base structure
        transformation: transformation matrix of sublattice
        symprec: float
        """
        frac_coords = structure.frac_coords
        rotations, translations = get_symmetry_operations(structure, symprec)
        converter = DerivativeMultiLatticeHash(transformation, frac_coords)
        return cls(
            frac_coords=frac_coords,
            rotations=rotations,
            translations=translations,
            converter=converter,
        )
